/* tripleInt.h
 */
#ifndef EVAL_CONTAINER_TRIPLE_INT_H
#define EVAL_CONTAINER_TRIPLE_INT_H
#include "osl/misc/carray.h"
#ifdef __INTEL_COMPILER
#  include <emmintrin.h>
#  define __builtin_ia32_pxor128 _mm_xor_pd
#  define __builtin_ia32_psubd128 _mm_sub_epi32
#  define __builtin_ia32_paddd128 _mm_add_epi32
#endif
#ifndef OSL_NO_SSE
#if (defined __x86_64) || (defined i386)
#  ifndef OSL_USE_SSE
#    define OSL_USE_SSE 1
#  endif
#else
#  warning "TripleInt without SSE"
#endif
#endif

#if (__GNUC__ >= 4 && __GNUC_MINOR__ >=4)
#  pragma GCC diagnostic warning "-Wuninitialized"
#endif

namespace osl
{
  namespace container
  {
#ifdef __INTEL_COMPILER
    typedef __m128i v4si;
    typedef __m128d v2di;
#else
    typedef int v4si __attribute__ ((vector_size (16)));
    typedef long long v2di __attribute__ ((vector_size (16)));
#endif
    struct TripleInt
    {
      union XMM{
	CArray<int,4> iv;
	v4si v4;
	v2di v2;
      } v __attribute__((aligned(16)));
      TripleInt(){
	clear();
      }
      TripleInt(TripleInt const& si){
#if OSL_USE_SSE
	v.v4=si.v.v4;
#else
	for(int i=0;i<3;i++) v.iv[i]=si.v.iv[i];
#endif
      }
      void clear()
      {
#if OSL_USE_SSE
	v.v2=__builtin_ia32_pxor128(v.v2,v.v2);
#else
	for(int i=0;i<3;i++) v.iv[i]=0;
#endif
      }
      int& operator[](int i) { 
	return v.iv[i]; 
      }
      const int& operator[](int i) const { 
	return v.iv[i]; 
      }
      TripleInt operator-() const{
	TripleInt ret;
#if OSL_USE_SSE
	ret.v.v4=__builtin_ia32_psubd128(ret.v.v4,v.v4);
#else
	for(int i=0;i<3;i++) ret.v.iv[i]= -v.iv[i];
#endif	  
	return ret;
      }
      TripleInt& operator+=(TripleInt const& si){
#if OSL_USE_SSE
	v.v4=__builtin_ia32_paddd128(v.v4,si.v.v4);
#else
	for(int i=0;i<3;i++) v.iv[i]+=si.v.iv[i];
#endif	  
	return *this;
      }
      TripleInt& operator-=(TripleInt const& si){
#if OSL_USE_SSE
	v.v4=__builtin_ia32_psubd128(v.v4,si.v.v4);
#else
	for(int i=0;i<3;i++) v.iv[i]-=si.v.iv[i];
#endif	  
	return *this;
      }
      static size_t size() { return 3; }
    };
    inline TripleInt operator+(TripleInt const& si0,TripleInt const& si1)
    {
      TripleInt ret(si0);
      ret+=si1;
      return ret;
    }
    inline TripleInt operator-(TripleInt const& si0,TripleInt const& si1)
    {
      TripleInt ret(si0);
      ret-=si1;
      return ret;
    }
    inline bool operator==(TripleInt const& l,TripleInt const& r)
    {
      for(int i=0;i<3;i++) 
	if (l[i] != r[i])
	  return false;
      return true;
    }
    
    class TripleIntPair{
      CArray<TripleInt,2> v;
    public:
      TripleIntPair() {}
      const TripleInt& operator[](int i) const{
	return v[i];
      }
      const TripleInt& operator[](Player pl) const{
	return v[pl];
      }
      TripleInt& operator[](int i){
	return v[i];
      }
      TripleInt& operator[](Player pl){
	return v[pl];
      }
      TripleIntPair& operator+=(TripleIntPair const& a){
	v[0]+=a.v[0];
	v[1]+=a.v[1];
	return *this;
      }
      TripleIntPair& operator-=(TripleIntPair const& a){
	v[0]-=a.v[0];
	v[1]-=a.v[1];
	return *this;
      }
    };
    inline TripleIntPair operator+(TripleIntPair const& si0,TripleIntPair const& si1)
    {
      TripleIntPair ret(si0);
      ret+=si1;
      return ret;
    }
    inline TripleIntPair operator-(TripleIntPair const& si0,TripleIntPair const& si1)
    {
      TripleIntPair ret(si0);
      ret-=si1;
      return ret;
    }
  }
  using container::TripleInt;
  using container::TripleIntPair;
#ifndef MINIMAL
  inline std::ostream& operator<<(std::ostream& os,TripleInt const& ti)
  {
    return os << "[" << ti[0] << "," << ti[1] << "," << ti[2] << "]";
  }
#endif
}
#endif // EVAL_CONTAINER_TRIPLE_INT_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
