#ifndef _PD_NTESUKI_TABLE_H
#define _PD_NTESUKI_TABLE_H
#include "osl/ntesuki/ntesukiRecord.h"
#include "osl/ntesuki/ntesukiMove.h"
#include "osl/misc/carray.h"
#include "osl/stl/hash_set.h"
#include "osl/pathEncoding.h"
#include "osl/hash/hashKey.h"
#include "osl/stl/hash_map.h"
#include "osl/stl/slist.h"
#include <boost/scoped_ptr.hpp>
#include <stdexcept>

namespace osl
{
  namespace stl
  {
    template <>
    struct hash<Position>{
      unsigned long operator()(const Position& p) const
      {
	return p.uintValue();
      }
    };
  }

  namespace ntesuki
  {
    /**
     * An exception thrown when the table is full.
     */
    struct TableFull : std::runtime_error
    {
      TableFull() : std::runtime_error("table full")
      {
      }
    };

    /**
     * An exception thrown when forEcachRecordFromRoot is called
     * althogh the root state is not set.
     */
    struct RootStateNotSet : std::runtime_error
    {
      RootStateNotSet () : std::runtime_error("root node not set")
      {
      }
    };
    /**
     * A table to hold ntesukiRecord.
     */
    class NtesukiTable
    {
    private:
      typedef hash_map<SignatureKey, NtesukiRecord::RecordList> ntesuki_hash_map;

    public:
      class Table : public ntesuki_hash_map
      {
      public:
	unsigned int capacity, default_gc_size;
	bool verbose, no_gc, gc_request;
	unsigned int numEntry, numCacheHit, gcCount;
	NtesukiRecord *root;
	boost::scoped_ptr<NumEffectState> rootState;
	static int largeGCCount;

	Table(unsigned int capacity,
	      unsigned int default_gc_size,
	      bool verbose);

	~Table();

	/**
	 * @c key б Record  Table õ.
	 * ⤷ϿƤʤä鿷Ͽ.
	 * @param key ̤ Hash
	 * @return б NtesukiRecord ؤΥݥ.
	 *         ( delete ƤϤʤʤ)
	 */
	NtesukiRecord *allocate(const HashKey& key,
				const PieceStand& white_stand,
				signed short distance);

	/**
	 * ɽõϿϤʤ
	 * @return ¸ߤʤ0
	 *   ǤʤǳݤؤΥݥ
	 *   (ְäƤ delete ʤ)
	 */
	NtesukiRecord *find(const HashKey& key);

	/**
	 * ɽϿ줿Ǥ.
	 */
	void erase(const HashKey key);

	/**
	 * ơ֥Ͽ줿 record  F ǽ.
	 */
	template <class F> void forEachRecord(F& f);
	template <class F> void forEachRecordFrom(F&,
						  NumEffectState&,
						  NtesukiRecord *);
	template <class F> void forEachRecordFromRoot(F& f);

	/**
	 * Collect garbage, until the size of the table reduces to
	 * @c gc_size
	 */
	void collectGarbage(unsigned int gc_size);
      };


    private:
      boost::scoped_ptr<Table> table;
      bool verbose;

    public:
      typedef NtesukiRecord record_t;

      struct HashPathEncoding
      {
	unsigned long operator()(PathEncoding const& pe) const
	{
	  return pe.getPath();
	}
      };
      typedef hash_set<PathEncoding, HashPathEncoding> PathSet;

      std::vector<int> depths;

      /**
       * @param capacity ɽݻ
       */
      NtesukiTable(unsigned int capacity,
		   unsigned int default_gc_size=0,
		   bool verbose=false);
      ~NtesukiTable();

      void clear()
      {
	table->clear();
      }

      Table::const_iterator begin() const
      {
	return table->begin();
      }
      Table::const_iterator end() const
      {
	return table->end();
      }

      /**
       * ơ֥Ҥ. ⤷ǤĤʤä硤
       * ơ֥礭䤷ɤΤʤ顤
       * Ǥ allocate .
       */
      NtesukiRecord *allocateRoot(const HashKey& key,
				  const PieceStand& white_stand,
				  signed short distance,
				  const NumEffectState* root_state = NULL)
      {
	table->root = table->allocate(key, white_stand, distance);
	if (root_state)
	{
	  table->rootState.reset(new NumEffectState(*root_state));
	}
	return table->root;
      }

      NtesukiRecord *allocateWithMove(NtesukiRecord* record,
				      const NtesukiMove& move)
      {
	/*  white stand ׻ƤΤ̵ */
	PieceStand white_stand = record->white_stand;
	const Move m = move.getMove();
	if (!move.isPass() && m.player())
	{
	  if (m.isDrop())
	  {
	    white_stand.sub(m.ptype());
	  }
	  else if (m.capturePtype() != PTYPE_EMPTY)
	  {
	    white_stand.add(unpromote(m.capturePtype()));
	  }
	}
	unsigned short child_distance = record->distance + 1;
	NtesukiRecord *child = table->allocate(record->key.newHashWithMove(m),
					       white_stand, child_distance);
	if (child)
	{
	  child->distance = std::min(child->distance, child_distance);
	  child->checkNewParent(record);
	}
	return child;
      }

      /**
       * ơ֥礭Ѳ find .
       */
      NtesukiRecord *find(const HashKey& key)
      {
	return table->find(key);
      }

      const NtesukiRecord *find(const HashKey& key) const
      {
	return table->find(key);
      }

      /**
       * ɽϿ줿Ǥ.
       */
      void erase(const HashKey key)
      {
	table->erase(key);
      }

      /**
       * ɽ.
       */
      void collectGarbage(unsigned int gc_size)
      {
	table->collectGarbage(gc_size);
      }

      /**
       * Ϳ줿 @param move ˳ǼƤ֤Ȥ.
       * ʤäơ֥뤫.
       */
      NtesukiRecord *findWithMove(NtesukiRecord *record,
				  const NtesukiMove& move)
      {
	/* ơ֥Ĵ١ä move ϿƤ
	 *
	 * (طʤȤäƤ move 
	 *  record ˤʤ piece  drop ǽ)
	 */
	if (move.isNormal() && move.isDrop())
	{
	  const Ptype drop_type = unpromote(move.getMove().ptype());
	  const PieceStand ps = record->getPieceStandSlow(move.getMove()
							  .player());
	  if (ps.get(drop_type) == 0)
	  {
	    return NULL;
	  }
	}
	NtesukiRecord *child = table->find(record->key.newHashWithMove(move.getMove()));
	if (child)
	{
	  child->checkNewParent(record);
	}
	return child;
      }

      NtesukiRecord *findWithMoveConst(const NtesukiRecord *record,
				       const NtesukiMove& move)
      {
	/* ơ֥Ĵ١ä move ϿƤ
	 *
	 * (طʤȤäƤ move 
	 *  record ˤʤ piece  drop ǽ)
	 */
	if (move.isNormal() && move.isDrop())
	{
	  const Ptype drop_type = unpromote(move.getMove().ptype());
	  const PieceStand ps = record->getPieceStandSlow(move.getMove()
							  .player());
	  if (ps.get(drop_type) == 0)
	  {
	    return NULL;
	  }
	}
	return table->find(record->key.newHashWithMove(move.getMove()));
      }

      /**
       * ơ֥Ͽ줿 record  F ǽ.
       */
      template <class F> void forEachRecord(F& f)
      {
	table->forEachRecord<F>(f);
      }

      /**
       * ơ֥ root node ֤Ĵ٤.
       */
      template <class F> void forEachRecordFromRoot(F& f)
      {
	table->forEachRecordFromRoot<F>(f);
      }

      /**
       * ơ֥Ͽ줿 record ο.
       */
      unsigned int size() const
      {
	return table->numEntry;/* not size() */
      }

      unsigned int capacity() const
      {
	return table->capacity;
      }

      void lockGC()
      {
	table->no_gc = true;
      }

      void unlockGC()
      {
	table->no_gc = false;
	if (table->gc_request && (table->default_gc_size > 0))
	{
	  table->collectGarbage(table->default_gc_size);
	  table->gc_request = false;
	}
      }

      bool isVerbose() const;
    };
  } //ntesuki
}// osl

#endif /* _PD_NTESUKI_TABLE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
