/* hasTimer.h
 */
#ifndef OSL_SEARCHTIMER_H
#define OSL_SEARCHTIMER_H
#include "osl/misc/milliSeconds.h"
#include "osl/misc/lightMutex.h"
#include "osl/oslConfig.h"
#include <boost/shared_ptr.hpp>

namespace osl
{
  namespace search
  {
    struct TimeAssigned
    {
      MilliSeconds::Interval standard, max;
      TimeAssigned() 
	: standard(MilliSeconds::Interval::infinity()),
	  max(MilliSeconds::Interval::infinity())
      {
      }
      explicit TimeAssigned(MilliSeconds::Interval assign) 
	: standard(assign), max(assign)
      {
      }
      TimeAssigned(MilliSeconds::Interval s, MilliSeconds::Interval m) 
	: standard(s), max(m)
      {
      }
    };
    struct SearchTimerCommon
    {
      /** õϻ */
      MilliSeconds start_time;
      /** õͽ */
      TimeAssigned assigned;
      /** ֤ܻĤäƤ鼡iteration ˿ʤफ */
      volatile double next_iteration_coefficient;
      volatile bool stop_all;
      
      // work area
      MilliSeconds last_tested;
      int next_node_count;
      double nps;
      volatile bool stable;
      typedef LightMutex Mutex;
      mutable Mutex mutex;

      SearchTimerCommon() 
	: start_time(MilliSeconds::now()), 
	  next_iteration_coefficient(4.0), stop_all(0), stable(true)
      {
      }
    };
    class SearchTimer
    {
      boost::shared_ptr<SearchTimerCommon> shared_timer;
      typedef SearchTimerCommon::Mutex Mutex;
    public:
      SearchTimer() : shared_timer(new SearchTimerCommon) {}
      SearchTimer(const SearchTimer& src) : shared_timer(src.shared_timer) {}
      virtual ~SearchTimer();
      void setTimeAssign(const TimeAssigned& a) {
	Mutex::scoped_lock lk(shared_timer->mutex);
	shared_timer->assigned = a;
      }
      void setStartTime(MilliSeconds start) {
	Mutex::scoped_lock lk(shared_timer->mutex);
	shared_timer->start_time = start;
	shared_timer->next_node_count = 0;
	shared_timer->nps = 0.0;
	shared_timer->last_tested = start;
	shared_timer->stop_all = false;
      }
      void setStable(bool new_stable) { shared_timer->stable = new_stable; }
      bool isStableNow() const { return shared_timer->stable; }
      bool hasSchedule() const {
	Mutex::scoped_lock lk(shared_timer->mutex);
	return ! shared_timer->assigned.standard.isInfinity();
      }
      const TimeAssigned& timeAssigned() const
      {
	Mutex::scoped_lock lk(shared_timer->mutex);
	return shared_timer->assigned;
      }
      const MilliSeconds startTime() const {
	Mutex::scoped_lock lk(shared_timer->mutex);
	return shared_timer->start_time;
      }
      double elapsed(MilliSeconds now) const 
      {
	return (now - shared_timer->start_time).toSeconds();
      }
      double elapsed() const { return elapsed(MilliSeconds::now()); }

      void setNextIterationCoefficient(double new_value) {
	Mutex::scoped_lock lk(shared_timer->mutex);
	shared_timer->next_iteration_coefficient = new_value;
      }
      double nextIterationCoefficient() const {
	Mutex::scoped_lock lk(shared_timer->mutex);
	return shared_timer->next_iteration_coefficient;
      }

      bool stopping() const { return shared_timer->stop_all; }
      void stopNow() { shared_timer->stop_all = true; }
      void throwIfNoMoreTime(int node_count)
      {
	// TODO: measure NPS
	SearchTimerCommon& shared = *shared_timer;
	if (! shared.stop_all) {
	  int next_node_count;
	  {
#ifdef OSL_USE_RACE_DETECTOR
	    Mutex::scoped_lock lk(shared_timer->mutex);
#endif
	    next_node_count = shared.next_node_count;
	  }
	  if (next_node_count > node_count || ! hasSchedule())
	    return;
	}
	testAndUpdateNextTimeTest(node_count);
      }
    private:
      void testAndUpdateNextTimeTest(int node_count);
    };
  } // namespace search
} // namespace osl


#endif /* OSL_SEARCHTIMER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
