/* SparseRegressionMultiplier.h
 */
#ifndef _STAT_SPARSEREGRESSIONMULTIPLIER_H
#define _STAT_SPARSEREGRESSIONMULTIPLIER_H

#include "osl/stat/matrixMultiplier.h"
#include "osl/stl/vector.h"

namespace osl
{
  namespace stat
  {
    /**
     * incrementaly compute out=Ap
     * where A = X^tX
     */
    class SparseRegressionMultiplier : public MatrixMultiplier
    {
      /** ĴŤߤ dimension.  οľܴʤ */
      unsigned int m_dimA;
      /** regularization term */
      double m_lambda;
      unsigned int m_lambda_start;
      int split;
    public:
      // CREATORS
      explicit SparseRegressionMultiplier(unsigned int dimA=0, double lambda=0.0, unsigned int lambda_start=0);
      ~SparseRegressionMultiplier();
      // MANIPULATORS
      void setSplit(int num_cpus=0) { split = num_cpus; }
    protected:
      void init(unsigned int new_dimA) { m_dimA = new_dimA; }
    public:
      // ACCESSORS
      /** return X^tX * p */
      void prod(const double *p, double *out) const;
      /** return X^t D X * p, where D is a diagonal matrix */
      void prodDiag(const double *d, const double *p, double *out) const;
      /** return X * q */
      void prodRaw(const double *q, double *out) const;
      unsigned int dim() const;
      /**
       * @param num_elements negative if the end of sequence (the next call must return
       *   the beginning of the sequence)
       */
      virtual void getVectorXWithID(unsigned int& instance_id,
				    int& num_elements,
				    unsigned int *non_zero_indices, 
				    double *non_zero_values) const =0;
      virtual void getVectorXWithIDPar(unsigned int& instance_id,
				       int& num_elements,
				       unsigned int *non_zero_indices, 
				       double *non_zero_values,
				       int tid) const;
      virtual int blockSize() const;
      virtual void getVectorXWithIDBlock(int& num_filled,
					 vector<unsigned int>& instance_id,
					 vector<int>& num_elements,
					 vector<vector<unsigned int> >& non_zero_indices, 
					 vector<vector<double> >& non_zero_values) const;
      virtual void newIteration() const;
    protected:
      /** partially compute X^tX*p and add to out where non_zero_{indices,values} reperesent one row in X */
      void addProduct(unsigned int non_zeros,
		      const unsigned int *non_zero_indices, 
		      const double *non_zero_values,
		      const double *p, double *out) const;
      void addProductDiag(unsigned int non_zeros,
			  const unsigned int *non_zero_indices, 
			  const double *non_zero_values,
			  const double d, const double *p, double *out) const;
    public:
      /**
       * @param xty is b such that Ax=b to solve x
       * @param diag_inv used in order to make diagonal preconditioner of X^tX
       * @return #instances
       */
      virtual unsigned int computeXtY(const double *y, double *xty, double *diag_inv) const;
      virtual unsigned int computeXtYDiag(const double *weight, const double *y, double *xty, double *diag_inv) const;

      struct ProdDiag;
      struct ProdDiagPar;
      struct ProdDiagBlock;
    };    

    struct DoubleReader
    {
      virtual ~DoubleReader() {}
      virtual double read() = 0;
    };
    struct DoubleConstReader : public DoubleReader
    {
      const double value;
      explicit DoubleConstReader(double v) : value(v)
      {
      }
      double read()
      {
	return value;
      }
    };
    struct DoubleReferenceReader : public DoubleReader
    {
      const double& value;
      explicit DoubleReferenceReader(double& v) : value(v)
      {
      }
      double read()
      {
	return value;
      }
    };
    struct DoubleArrayReader : public DoubleReader
    {
      const double *array;
      unsigned int cur;
      explicit DoubleArrayReader(const double *a) : array(a), cur(0)
      {
      }
      double read()
      {
	return array[cur++];
      }
    };
    template <class Iter>
    struct DoubleIteratorReader : public DoubleReader
    {
      Iter it;
      explicit DoubleIteratorReader(Iter i) : it(i)
      {
      }
      double read()
      {
	return *it++;
      }
    };

    /** getVectorX() ̤ɤ߹Ȳꤷ*/
    class SparseRegressionMultiplierSeq : public SparseRegressionMultiplier
    {
      mutable unsigned int current_id;
    public:
      // CREATORS
      explicit SparseRegressionMultiplierSeq(unsigned int dimA=0, double lambda=0.0, unsigned int lambda_start=0);
      ~SparseRegressionMultiplierSeq();
      // MANIPULATORS
    public:
      void getVectorXWithID(unsigned int& instance_id,
			    int& num_elements,
			    unsigned int *non_zero_indices, 
			    double *non_zero_values) const;
      virtual void getVectorX(int& num_elements,
			      unsigned int *non_zero_indices, 
			      double *non_zero_values) const =0;
    };

    /**
     * feature ͤ 1.0 ˸ꤵƤС
     */
    class BinarySparseRegressionMultiplier : public MatrixMultiplier
    {
      unsigned int m_dimA;
      double m_lambda;
    public:
      // CREATORS
      explicit BinarySparseRegressionMultiplier(unsigned int dimA,
						double lambda=0.0);
      ~BinarySparseRegressionMultiplier();
      // MANIPULATORS
      // ACCESSORS
      void prod(const double *p, double *out) const;
      unsigned int dim() const;
      virtual void getVectorX(int& num_elements,
			      unsigned int *non_zero_indices) const =0;
      virtual void newIteration() const;
      /** partially compute X^tX*p and add to out where non_zero_indices reperesent one row in X */
      void addProduct(unsigned int non_zeros,
		      const unsigned int *non_zero_indices, 
		      const double *p, double *out) const;
    };    
    
  } // namespace stat
} // namespace osl


#endif /* _STAT_SPARSEREGRESSIONMULTIPLIER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
