/* dualDfpn.cc
 */
#include "osl/checkmate/dualDfpn.h"
#include "osl/checkmate/dfpn.h"
#include "osl/repetitionCounter.h"
#include "osl/stl/hash_map.h"
#include "osl/stat/average.h"
#include "osl/centering3x3.h"
#include "osl/centering5x3.h"
#ifdef OSL_SMP
#  include "osl/misc/lightMutex.h"
#endif
#ifdef OSL_SHOW_PROOF_TREE_MIGRATION_STAT
#  include "osl/stat/ratio.h"
#endif
#include <iostream>
#define DFPN_SHARE_TABLE

static const int max_oracle_list_size = 2;
static const size_t local_table_growth_limit = 40000;
struct osl::checkmate::DualDfpn::OraclePool
{
  struct Element
  {
    Dfpn::ProofOracle oracle;
    PieceStand proof_pieces;
    unsigned int id;
    bool in_check;
    Element() : oracle(HashKey(), PieceStand()), id((unsigned int)-1), in_check(false)
    {
    }
    Element(const Dfpn::ProofOracle& o, PieceStand p, size_t i, bool c) : oracle(o), proof_pieces(p), id(i), in_check(c)
    {
    }
  };
  struct List : FixedCapacityVector<Element, max_oracle_list_size>
  {
    void add(const Element& e)
    {
      if (size() == capacity())
	back() = e;
      else
	push_back(e);
    }
  };
#ifdef OSL_SMP
  typedef osl::misc::LightMutex Mutex;
  mutable Mutex mutex;
#endif  
  typedef hash_map<HashKey, List> table_t;
  table_t table;
  Player defender;
  void setAttack(Player attack) 
  {
    defender = alt(attack);
  }
  void addProof(const NumEffectState& state, const HashKey& key, PieceStand proof_pieces) 
  {
    const Dfpn::ProofOracle oracle(key, PieceStand(WHITE, state));
    const std::pair<HashKey,HashKey> king = makeLargeKey(state);
#ifdef OSL_SMP
    Mutex::scoped_lock lk(mutex);
#endif
    const Element e(oracle, proof_pieces, table.size(), state.inCheck());
    table[king.first].add(e);
    table[king.second].add(e);
  }
  const List probe(const NumEffectState& state) const
  {
    const std::pair<HashKey,HashKey> key = makeLargeKey(state);
#ifdef OSL_SMP
    Mutex::scoped_lock lk(mutex);
#endif
    table_t::const_iterator p = table.find(key.first);
    if (p != table.end())
      return p->second;
    p = table.find(key.second);
    if (p != table.end())
      return p->second;
    return List();
  }

  template <Direction DIR>
  static void addKey(HashKey& key, const SimpleState& state, Position target)
  {
    const Offset offset = DirectionTraits<DIR>::blackOffset();
    target += offset;	// 8 ˵ƻʤ֤ˤѴ
    const Piece piece = state.getPieceOnBoard(target);
    hash::Hash_Gen_Table.addHashKey(key, target, piece.ptypeO());
  }
  template <Direction DIR, Direction DIR2>
  static void addKey(HashKey& key, const SimpleState& state, Position target)
  {
    const Offset offset = DirectionTraits<DIR>::blackOffset()
      + DirectionTraits<DIR2>::blackOffset();
    target += offset;
    const Piece piece = state.getPieceOnBoard(target);
    hash::Hash_Gen_Table.addHashKey(key, target, piece.ptypeO());
  }
  const HashKey makeKey(const SimpleState& state) const
  {
    const Position target_king=state.getKingPosition(defender);
    const Position center = Centering3x3::adjustCenter(target_king);
    HashKey key;
    hash::Hash_Gen_Table.addHashKey(key, center, 
				    state.getPieceOnBoard(center).ptypeO());
    addKey<UL>(key, state, center); addKey<U> (key, state, center);
    addKey<UR>(key, state, center);
    addKey<L> (key, state, center); addKey<R> (key, state, center);
    addKey<DL>(key, state, center); addKey<D> (key, state, center);
    addKey<DR>(key, state, center);
    return key;
  }
  const std::pair<HashKey,HashKey> makeLargeKey(const SimpleState& state) const
  {
    HashKey key_small = makeKey(state), key_large;
    const Position target_king=state.getKingPosition(defender);
    const Position center = Centering5x3::adjustCenter(target_king);
    hash::Hash_Gen_Table.addHashKey(key_large, center, 
				    state.getPieceOnBoard(center).ptypeO());
    addKey<UL>(key_large, state, center); addKey<U> (key_large, state, center);
    addKey<UR>(key_large, state, center);
    addKey<L> (key_large, state, center); addKey<R> (key_large, state, center);
    addKey<DL>(key_large, state, center); addKey<D> (key_large, state, center);
    addKey<DR>(key_large, state, center);
    addKey<L,UL>(key_large, state, center); addKey<L,L> (key_large, state, center);
    addKey<L,DL>(key_large, state, center);
    addKey<R,UR>(key_large, state, center); addKey<R,R> (key_large, state, center);
    addKey<R,DR>(key_large, state, center);
    return std::make_pair(key_large, key_small);
  }
};

struct osl::checkmate::DualDfpn::Shared
{
  CArray<DfpnTable,2> table;
  CArray<OraclePool,2> pool;
  size_t main_node_count;
  size_t simulation_count;
  CArray<stat::Average,max_oracle_list_size> proof_by_oracle;
  CArray<bool,2> blocking_verify;
#ifdef OSL_SMP
  LightMutex mutex;
  CArray<LightMutex,max_oracle_list_size> proof_by_oracle_mutex;
#endif
  Shared() : main_node_count(0), simulation_count(0)
  {
    table[BLACK].setAttack(BLACK);
    table[WHITE].setAttack(WHITE);
    pool[BLACK].setAttack(BLACK);
    pool[WHITE].setAttack(WHITE);
    blocking_verify.fill(true);
  }
  ~Shared()
  {
    showStats();
  }
  void showStats()
  {
    if (main_node_count || simulation_count) {
#ifdef DFPN_DEBUG
      std::cerr << "shared " << main_node_count << " " << simulation_count << "\n";
      for (size_t i=0; i<proof_by_oracle.size(); ++i)
	std::cerr << proof_by_oracle[i].getAverage() 
		  << " " << (int)(proof_by_oracle[i].getAverage()*proof_by_oracle[i].numElements()) << "\n";
      std::cerr << "oracles " << pool[BLACK].table.size() << " " << pool[WHITE].table.size() << "\n";
      std::cerr << "table " << table[0].totalSize() << " " << table[1].totalSize() << "\n";
      table[0].showStats();
      table[1].showStats();
#endif
    }
  }
  void addMainNodeCount(int add)
  {
#ifdef OSL_SMP
    LightMutex::scoped_lock lk(mutex);
#endif
    main_node_count += add;
  }
  void addSimulationNodeCount(int add)
  {
#ifdef OSL_SMP
    LightMutex::scoped_lock lk(mutex);
#endif
    simulation_count += add;
  }
};

struct osl::checkmate::DualDfpn::Local
{
  CArray<Dfpn,2> dfpn, dfpn_small;
#ifndef DFPN_SHARE_TABLE
  CArray<DfpnTable,2> table;
#endif
  CArray<DfpnTable,2> table_small;
  size_t local_node_count;
  Local() : local_node_count(0)
  {
#ifndef DFPN_SHARE_TABLE
    table[BLACK].setAttack(BLACK);
    table[WHITE].setAttack(WHITE);
#endif
    table_small[BLACK].setAttack(BLACK);
    table_small[WHITE].setAttack(WHITE);
  }
  ~Local()
  {
#ifdef DFPN_DEBUG
    std::cerr << "local " << table_small[0].totalSize()
	      << " " << table_small[1].totalSize() << "\n";
#endif
  }
};

/* ------------------------------------------------------------------------- */

osl::checkmate::
DualDfpn::DualDfpn(size_t /*limit*/)
  : shared(new Shared), local(new Local)
{
#ifdef DFPN_SHARE_TABLE
  local->dfpn[BLACK].setTable(&(shared->table[BLACK]));
  local->dfpn[WHITE].setTable(&(shared->table[WHITE]));
#else
  local->dfpn[BLACK].setTable(&(local->table[BLACK]));
  local->dfpn[WHITE].setTable(&(local->table[WHITE]));
#endif
  local->dfpn_small[BLACK].setTable(&(local->table_small[BLACK]));
  local->dfpn_small[WHITE].setTable(&(local->table_small[WHITE]));

  updateBlockingVerify();
}

osl::checkmate::
DualDfpn::DualDfpn(const DualDfpn& src)
  : shared(src.shared), local(new Local)
{
#ifdef DFPN_SHARE_TABLE
  local->dfpn[BLACK].setTable(&(shared->table[BLACK]));
  local->dfpn[WHITE].setTable(&(shared->table[WHITE]));
#else
  local->dfpn[BLACK].setTable(&(local->table[BLACK]));
  local->dfpn[WHITE].setTable(&(local->table[WHITE]));
#endif
  local->dfpn_small[BLACK].setTable(&(local->table_small[BLACK]));
  local->dfpn_small[WHITE].setTable(&(local->table_small[WHITE]));

  updateBlockingVerify();
}

void osl::checkmate::
DualDfpn::updateBlockingVerify()
{
  local->dfpn[BLACK].setBlockingVerify(shared->blocking_verify[BLACK]);
  local->dfpn[WHITE].setBlockingVerify(shared->blocking_verify[WHITE]);
  local->dfpn_small[BLACK].setBlockingVerify(shared->blocking_verify[BLACK]);
  local->dfpn_small[WHITE].setBlockingVerify(shared->blocking_verify[WHITE]);
}

osl::checkmate::
DualDfpn::~DualDfpn()
{
}

template <osl::Player P>
bool
#if (defined __GNUC__) && (! defined GPSONE)
__attribute__ ((noinline))
#endif
 osl::checkmate::
DualDfpn::isWinningState(int node_limit, NumEffectState& state, 
			 const HashKey& key, const PathEncoding& path,
			 Move& best_move, Move last_move)
{
  assert(state.getTurn() == P);
  // oracle
  Dfpn& dfpn = local->dfpn[P];
  const OraclePool::List l = shared->pool[P].probe(state);
  const PieceStand attack_stand = (P==BLACK) ? key.blackStand() : PieceStand(WHITE, state);
  int num_tried = 0;
  for (size_t i=0; i<l.size(); ++i)
  {
    if (! attack_stand.isSuperiorOrEqualTo(l[i].proof_pieces)
	|| l[i].in_check != state.inCheck())
      continue;
    ++num_tried;
    const ProofDisproof pdp = (node_limit > 20)
      ? dfpn.tryProof(state, key, path, l[i].oracle, l[i].id, best_move, last_move)
      : dfpn.tryProofLight(state, key, path, l[i].oracle, l[i].id, best_move, last_move);
    const size_t count = dfpn.nodeCount();
    local->local_node_count += count;
    shared->addSimulationNodeCount(count);
    if (count) {
#ifdef OSL_SMP
      LightMutex::scoped_lock lk(shared->proof_by_oracle_mutex[i]);
#endif
      shared->proof_by_oracle[i].add(pdp.isCheckmateSuccess());
    }
    if (pdp.isFinal())
      return pdp.isCheckmateSuccess();
  }
  if (node_limit == 0 && num_tried) 
    return false;			// already tested table
  const ProofDisproof table_pdp = dfpn.hasCheckmateMove(state, key, path, 0, best_move, last_move);
  if (table_pdp.isCheckmateSuccess())
    return true;
#ifdef OSL_SHOW_PROOF_TREE_MIGRATION_STAT
  static stat::Ratio migration_success("migration_success", true);
  bool need_migration = false;
#endif
  // local
  if (node_limit < 80) {
    if (local->table_small[P].totalSize() >= local_table_growth_limit) {
      local->table_small[P].clear();
    }
    Dfpn& dfpn_small = local->dfpn_small[P];
    const ProofDisproof pdp = dfpn_small.hasCheckmateMove(state, key, path, node_limit, best_move, last_move);
    const size_t count = dfpn_small.nodeCount();
    local->local_node_count += count;
    shared->addMainNodeCount(count);
    if (! pdp.isCheckmateSuccess())
      return false;
    // fall through if checkmate success (TODO: efficient proof tree migration)
#ifdef OSL_SHOW_PROOF_TREE_MIGRATION_STAT
    need_migration = true;
#endif
  }
  // main
  PieceStand proof_pieces;
  const ProofDisproof pdp = dfpn.hasCheckmateMove(state, key, path, node_limit, best_move, proof_pieces, last_move);
  const size_t count = dfpn.nodeCount();
  local->local_node_count += count;
  shared->addMainNodeCount(count);
  if (pdp.isCheckmateSuccess())
    shared->pool[P].addProof(state, key, proof_pieces);
#ifdef OSL_SHOW_PROOF_TREE_MIGRATION_STAT
  if (need_migration)
    migration_success.add(pdp.isCheckmateSuccess());
#endif
  return pdp.isCheckmateSuccess();
}

bool osl::checkmate::
DualDfpn::isWinningState(int node_limit, NumEffectState& state, 
			   const HashKey& key, const PathEncoding& path,
			   Move& best_move, Move last_move)
{
  if (state.getTurn() == BLACK)
    return isWinningState<BLACK>(node_limit, state, key, path, best_move, last_move);
  else
    return isWinningState<WHITE>(node_limit, state, key, path, best_move, last_move);
}

template <osl::Player P>
bool
#if (defined __GNUC__) && (! defined GPSONE)
__attribute__ ((noinline))
#endif
 osl::checkmate::
DualDfpn::isLosingState(int node_limit, NumEffectState& state, 
			const HashKey& key, const PathEncoding& path,
			Move last_move)
{
  assert(state.getTurn() == P);
  Dfpn& dfpn = local->dfpn[alt(P)];
  const ProofDisproof pdp = dfpn.hasEscapeMove(state, key, path, node_limit, last_move);
  const size_t count = dfpn.nodeCount();
  local->local_node_count += count;
  shared->addMainNodeCount(count);
  return pdp.isCheckmateSuccess();
}

bool osl::checkmate::
DualDfpn::isLosingState(int node_limit, NumEffectState& state, 
			  const HashKey& key, const PathEncoding& path,
			  Move last_move)
{
  if (state.getTurn() == BLACK)
    return isLosingState<BLACK>(node_limit, state, key, path, last_move);
  else
    return isLosingState<WHITE>(node_limit, state, key, path, last_move);
}

void osl::checkmate::
DualDfpn::writeRootHistory(const RepetitionCounter& counter,
			   const MoveStack& moves,
			   const SimpleState& /*state*/, Player attack)
{
  // TODO: ɽ򥯥ꥢƤޤ˺ => DualDfpn ǵ뤷ɤ
  Dfpn& dfpn = local->dfpn[attack];
  for (int i=0; i<counter.checkCount(attack); ++i)
  {
    const HashKey& key = counter.history().top(i);
    if (key != counter.history().top(0)) // ignore current state
    {
      dfpn.setIllegal(key);
    }
    assert(moves.hasLastMove(i+1)); // oops, different index
    if (! moves.hasLastMove(i+1))
      break;
  }
}

void osl::checkmate::
DualDfpn::setRootPlayer(Player root)
{
  shared->blocking_verify[root] = true;
  shared->blocking_verify[alt(root)] = false;

  updateBlockingVerify();
}

void osl::checkmate::
DualDfpn::setVerbose(int /*level*/)
{
}

int osl::checkmate::
DualDfpn::distance(Player attack, const HashKey& key) const
{
  return local->dfpn[attack].distance(key);
}

size_t osl::checkmate::
DualDfpn::mainNodeCount() const
{
#ifdef OSL_USE_RACE_DETECTOR
  LightMutex::scoped_lock lk(shared->mutex);
#endif
  return shared->main_node_count;
  // return shared->table[BLACK].totalSize() + shared->table[WHITE].totalSize();
}

size_t osl::checkmate::
DualDfpn::totalNodeCount() const
{
#ifdef OSL_USE_RACE_DETECTOR
  LightMutex::scoped_lock lk(shared->mutex);
#endif
  return shared->main_node_count + shared->simulation_count;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
