/* moveGenerator.cc
 */
#include "osl/search/moveGenerator.h"
#include "osl/search/searchState2.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/search/sortCaptureMoves.h"
#include "osl/search/breakThreatmate.h"
#include "osl/search/analyzer/categoryMoveVector.h"
#include "osl/move_generator/capture_.h"
#include "osl/move_generator/escape_.h"
#include "osl/move_generator/promote_.h"
#include "osl/move_generator/addEffect_.h"
#include "osl/move_generator/allMoves.h"
#include "osl/move_generator/attackToPinned.h"
#include "osl/move_action/store.h"
#include "osl/move_classifier/check_.h"
#include "osl/move_classifier/safeMove.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/rating/featureSet.h"
#include "osl/rating/ratingEnv.h"
#include "osl/eval/pieceEval.h"
#include "osl/eval/progressEval.h"
#include "osl/eval/experimental/testEval.h"
#include "osl/stat/average.h"
#include <iostream>
#include <iomanip>

// #define STAT_WIDTH_VS_LIMIT

#ifndef NDEBUG
#  define SAFE_MOVE_ONLY
#endif
const int max_see = 20000;

static const osl::rating::FeatureSet *static_feature_set;
static const osl::rating::FeatureSet& feature_set()
{
  return *static_feature_set;
}

void osl::search::MoveGenerator::initOnce()
{
  if (static_feature_set == 0)
    static_feature_set = &rating::StandardFeatureSet::instance();
}

namespace osl
{
  namespace search
  {
    const CArray2d<MoveGenerator::generator_t, 2, MoveGenerator::FINISH> MoveGenerator::Generators = {{
      {
	0,
	&MoveGenerator::generateKingEscape<BLACK>,
	&MoveGenerator::generateTakeBack<BLACK>,
	&MoveGenerator::generateBreakThreatmate<BLACK>,
	&MoveGenerator::generateCapture<BLACK>,
	0,
	&MoveGenerator::generateTesuji<BLACK>,
	&MoveGenerator::generateAll<BLACK>,
      },
      {
	0,
	&MoveGenerator::generateKingEscape<WHITE>,
	&MoveGenerator::generateTakeBack<WHITE>,
	&MoveGenerator::generateBreakThreatmate<WHITE>,
	&MoveGenerator::generateCapture<WHITE>,
	0,
	&MoveGenerator::generateTesuji<WHITE>,
	&MoveGenerator::generateAll<WHITE>,
      }
    }};
    const CArray<const char *, MoveGenerator::FINISH> MoveGenerator::GeneratorNames = {{
      "INITIAL", "KING_ESCAPE", "TAKEBACK", "BREAK_THREATMATE", "TACTICAL", "SENTINEL", "TESUJI", "ALL", 
    }};
#ifdef STAT_WIDTH_VS_LIMIT
    struct WidthVSLimit 
    {
      CArray<stat::Average,10> averages;

      ~WidthVSLimit()
      {
	report();
      }
      stat::Average& average(int limit) 
      {
	limit /= 100 - 3;
	return averages[std::min(std::max(limit,0),(int)averages.size()-1)];
      }
      void report()
      {
	std::cerr << "WidthVSLimit@MoveGenerator\n";
	for (int limit=300; limit<300+(int)averages.size()*100; limit+=100) {
	  std::cerr << std::setw(5) << limit << " " << average(limit).getAverage() << std::endl;
	}
      }
    } Width_VS_Limit;
#endif
    template
    void MoveGenerator::init<osl::eval::ProgressEval>(
      int limit, SimpleHashRecord *record, const osl::eval::ProgressEval&,
      const NumEffectState&, bool in_pv, Move hash_move, bool quiesce);
    template
    void MoveGenerator::init<osl::eval::experimental::TestEval>(
      int limit, SimpleHashRecord *record,
      const osl::eval::experimental::TestEval&,
      const NumEffectState&, bool in_pv, Move hash_move, bool quiesce);
  }
}

/* ------------------------------------------------------------------------- */

osl::search::
MoveMarker::MoveMarker() : cur(1) 
{
  marker.fill(0);
}

void osl::search::
MoveMarker::clear() 
{
  ++cur;
  if (cur == 0) {
    marker.fill(0);
    cur = 1;
  }
}

bool osl::search::
MoveMarker::registerIfNew(const NumEffectState& state, Move m)
{
  value_t& val = marker(toIndex(m), pieceIndex(state, m));
  if (val == cur)
    return false;
  val = cur;
  return true;
}

bool osl::search::
MoveMarker::registered(const NumEffectState& state, Move m) const
{
  return marker(toIndex(m), pieceIndex(state, m)) == cur;
}

/* ------------------------------------------------------------------------- */

osl::search::
MoveGenerator::MoveGenerator() : record(0), progress(16)
{
}

int osl::search::
MoveGenerator::captureValue(Ptype ptype)
{
  if (! isPiece(ptype))
    return 0;
  int result 
    = eval::Ptype_Eval_Table.captureValue(newPtypeO(WHITE, ptype));
  assert(result >= 0);
  return result;
}

template <class EvalT>
void osl::search::
MoveGenerator::init(int l, SimpleHashRecord *r,
		    const EvalT& eval,
		    const NumEffectState& state, bool in_pv, Move hash_move,
		    bool quiesce) 
{
  assert(r);
  assert(l > 0);
  limit = l;
  record = r;
  cur_state = INITIAL;
  moves.clear();
  cur_index = tried = 0;
  progress = eval.progress32();

  marker.clear();
  env.make(state, eval.pins(state.getTurn()), eval.pins(alt(state.getTurn())),
	   eval.progress16());
  if (hash_move.isNormal())
    marker.registerMove(state, hash_move);
  in_quiesce = quiesce;
  this->in_pv = in_pv;
}

void osl::search::
MoveGenerator::dump() const
{
  std::cerr << "generator " << cur_state << " index " << cur_index 
	    << " limit " << limit << " tried " << tried << "\n";
  std::cerr << moves.size() << "\n"
#ifndef MINIMAL
   << moves
#endif
    ;
}

template <osl::Player P>
const osl::MoveLogProb osl::search::
MoveGenerator::nextTacticalMoveWithGeneration(const SearchState2& state) 
{	
  assert(record);
  while (true) {
    assert(cur_index >= moves.size());
    if (cur_state == KING_ESCAPE && record->is_king_in_check) {
      cur_state = FINISH;
      break;
    }
    if (++cur_state >= TACTICAL_FINISH)
      break;
    // generate
    assert(Generators[playerToIndex(P)][cur_state]);
    (this->*Generators[playerToIndex(P)][cur_state])(state);
    if (cur_index < moves.size()) {
      ++tried;
      return moves[cur_index++];
    } 
  }
  return MoveLogProb();
}

template <osl::Player P>
const osl::MoveLogProb osl::search::
MoveGenerator::nextMoveWithGeneration(const SearchState2& state) 
{	
  assert(record);
  while (true) {
    assert(cur_index >= moves.size());
    if (++cur_state >= FINISH)
      break;
    // generate
    assert(Generators[playerToIndex(P)][cur_state]);
    (this->*Generators(playerToIndex(P), cur_state))(state);
    if (cur_index < moves.size()) {
      ++tried;
      return moves[cur_index++];
    } 
  }
  return MoveLogProb();
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateKingEscape(const SearchState2& sstate)
{
  env.history = sstate.history();
  if (! record->is_king_in_check) 
    return;
  using namespace move_action;

  const NumEffectState& state = sstate.state();
  const Piece king = state.getKingPiece<P>();
  assert(state.hasEffectBy(alt(P), king.position()));

  MoveVector src;
  Store store(src);
  move_generator::Escape<P,Store>::
    generateMoves(state,king,store);

  if (src.size() == 1) {
    moves.push_back(MoveLogProb(src[0], 20));
    return;
  }
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    const int prob = std::min(limit, feature_set().logProbKingEscape(state, env, *p));
    assert(prob > 0);
    moves.push_back(MoveLogProb(*p, prob));
  }
  moves.sortByProbability();
}



template <osl::Player P>
void osl::search::
MoveGenerator::generateBreakThreatmate(const SearchState2& sstate)
{
  const NumEffectState& state = sstate.state();
  const Move threatmate_move = record->threatmate().threatmateMove(state.getTurn());
  if (! threatmate_move.isNormal())
    return;
  BreakThreatmate::generate(limit, state, threatmate_move, moves);
  for (size_t i=0; i<moves.size(); ++i)
    marker.registerMove(state, moves[i].getMove());
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateTakeBack(const SearchState2& sstate)
{
  using namespace move_action;
  const Move last_move = sstate.lastMove();
  if (! last_move.isNormal())
    return;
  const Position last_to = last_move.to();
  
  const NumEffectState& state = sstate.state();
  MoveVector src;
  Store store(src);

  if (in_quiesce)
    return quiesceCapture<P>(state, last_to);

  move_generator::Capture<P>::generate(state, last_to, store);
  assert(moves.empty());
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    if (ShouldPromoteCut::canIgnoreMove<P>(*p))
      continue;
    const int prob = feature_set().logProbTakeBack(state, env, *p);
#ifdef OSL_SMP
    if (! p->isDrop() && p->ptype() != KING
	&& env.my_pin.test(state.getPieceOnBoard(p->from()).number())) {
      if (move_classifier::KingOpenMove<P>::isMember(state, p->ptype(), p->from(), p->to()))
	continue;
    }
#endif
    if (prob <= std::min(200, limit) && marker.registerIfNew(state, *p))
      moves.push_back(MoveLogProb(*p, prob));
  }
  moves.sortByProbability();
}

namespace osl
{
  template <Player P, Ptype PTYPE>
  static void makeCapture(const NumEffectState& state, 
				       MoveVector& out)
  {
    move_action::Store store(out);
    mask_t pieces = state.getOnBoardMask(alt(P)).selectBit<PTYPE>() 
      & state.effectedMask(P).getMask(PtypeFuns<PTYPE>::indexNum);
    while (pieces.any())
    {
      const Piece p = state.getPieceOf(pieces.takeOneBit()+PtypeFuns<PTYPE>::indexNum*32);
      assert(p.isOnBoardByOwner<PlayerTraits<P>::opponent>());
      move_generator::Capture<P>::generate(state, p.position(), store);
    }
  }
}

template <osl::Player P>
void osl::search::
MoveGenerator::addCapture(const NumEffectState& state, const RatingEnv& env, const MoveVector& src)
{
  if (in_quiesce) {
    for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
      if (ShouldPromoteCut::canIgnoreMove<P>(*p))
	continue;
      const int see = PieceEval::computeDiffAfterMoveForRP(state, *p);
      if (see < 0)
	continue;
      moves.push_back(MoveLogProb(*p, max_see - see));
    }
    return;
  }
  
  for (MoveVector::const_iterator p=src.begin(); p!=src.end(); ++p) {
    const Move m = *p;
    if (ShouldPromoteCut::canIgnoreMove<P>(m))
      continue;
#ifdef SAFE_MOVE_ONLY
    if (! m.isDrop() && m.ptype() != KING
	&& env.my_pin.test(state.getPieceOnBoard(m.from()).number())) {
      if (move_classifier::KingOpenMove<P>::isMember(state, m.ptype(), m.from(), m.to()))
	continue;
    }
#endif
    const int prob = feature_set().logProbSeePlus(state, env, *p);
    // const int prob = feature_set().logProbTakeBack(state, env, *p);
    if (prob <= 200 && marker.registerIfNew(state, *p)) {
      moves.push_back(MoveLogProb(*p, prob));
    }
  }
  return;
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateCapture(const SearchState2& sstate)
{
  using namespace move_action;

  const NumEffectState& state = sstate.state();
  MoveVector src;
#if 1
  // lance, bishop, rook
  makeCapture<P,LANCE>(state, src);
  makeCapture<P,BISHOP>(state, src);
  makeCapture<P,ROOK>(state, src);
  // knight, silver, gold
  makeCapture<P,KNIGHT>(state, src);
  makeCapture<P,SILVER>(state, src);
  makeCapture<P,GOLD>(state, src);
#else
  makeCaptureOtherThanPawn<P>(state, src);
#endif
  addCapture<P>(state, env, src);
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateTesuji(const SearchState2& sstate)
{
  const NumEffectState& state = sstate.state();
  if (in_quiesce) {
    MoveVector src;
    move_action::Store store(src);
    move_generator::Promote<P,move_action::Store>::
      generateMoves(state, store);
    makeCapture<P,PAWN>(state, src);
    addCapture<P>(state, env, src);
  }
}

template <osl::Player P>
void osl::search::
MoveGenerator::quiesceCapture(const NumEffectState& state, Position to)
{
  MoveVector moves;
  move_action::Store store(moves);
  move_generator::Capture<P>::generate(state, to, store);
  for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p) {
    if (ShouldPromoteCut::canIgnoreAndNotDrop<P>(*p))
      continue;
    int see = PieceEval::computeDiffAfterMoveForRP(state, *p);
    if (see < 0)
      continue;
    this->moves.push_back(MoveLogProb(*p, max_see - see));
  }
  this->moves.sortByProbabilityReverse();
}

template <osl::Player P>
void osl::search::
MoveGenerator::generateAll(const SearchState2& sstate)
{
  if (in_quiesce) 
    return;
  const NumEffectState& state = sstate.state();
  MoveLogProbVector all;
  feature_set().generateLogProb(state, env, limit, all, in_pv);
#ifdef STAT_WIDTH_VS_LIMIT
  const size_t moves_size_before = moves.size();
#endif
  for (size_t i=0; i<all.size(); ++i) {
    const Move m = all[i].getMove();
    if (ShouldPromoteCut::canIgnoreAndNotDrop<P>(m))
      continue;
    int limit = all[i].getLogProb();
    if (this->limit >= 400) {
      using namespace move_classifier;
      if (all[i].getMove().capturePtype() != PTYPE_EMPTY
	  || all[i].getMove().isPromote()
	  || (in_pv && MoveAdaptor<Check<P> >::isMember(state, all[i].getMove())))
	limit = std::min(limit, 400);
    }
    if (limit <= this->limit
	&& marker.registerIfNew(state, m)) {
#ifndef NDEBUG
      if (! m.isDrop()) {
	assert(! (env.my_pin.test(state.getPieceOnBoard(m.from()).number())
		  && move_classifier::KingOpenMove<P>::isMember(state, m.ptype(), m.from(), m.to())));
	assert(! (m.ptype() == KING && state.hasEffectBy(alt(P), m.to())));
      }
#endif
      moves.push_back(MoveLogProb(all[i].getMove(), limit));
    }
  }
#ifdef STAT_WIDTH_VS_LIMIT
  Width_VS_Limit.average(limit).add(moves.size() - moves_size_before);
#endif
}

#ifndef MINIMAL
void osl::search::
MoveGenerator::generateAll(Player P, const SearchState2& state, 
			   analyzer::CategoryMoveVector& out)
{
  assert(moves.size() == 0);

  for (int i=0; i<FINISH; ++i) {
    if (! Generators[playerToIndex(P)][i])
      continue;
    (this->*Generators[playerToIndex(P)][i])(state);
    out.push_front(analyzer::CategoryMoves(moves, GeneratorNames[i]));
    bool generated = moves.size();
    moves.clear();    
    if (i == KING_ESCAPE && generated)
      break;
  }
  out.reverse();
}
#endif

template <osl::Player P>
void
#if (defined __GNUC__) && (! defined GPSONE)
__attribute__ ((used))
#endif
  osl::search::
MoveGenerator::generateAll(const SearchState2& state, MoveLogProbVector& out) 
{
  using namespace move_classifier;
  for (MoveLogProb m = nextTacticalMove<P>(state);
       m.validMove(); m = nextTacticalMove<P>(state)) {
    if (ConditionAdaptor<SafeMove>::isMember(state.state(), m.getMove()))
      out.push_back(m);
  }
  for (MoveLogProb m = nextMove<P>(state); m.validMove(); 
       m = nextMove<P>(state)) {
    if (ConditionAdaptor<SafeMove>::isMember(state.state(), m.getMove()))
      out.push_back(m);
  }
}

void osl::search::
MoveGenerator::generateAll(Player P, const SearchState2& state, MoveLogProbVector& out) 
{
  if (P==BLACK)
    generateAll<BLACK>(state, out);
  else
    generateAll<WHITE>(state, out);
}

namespace osl
{
  namespace search
  {
    template const MoveLogProb MoveGenerator::nextMoveWithGeneration<BLACK>(const SearchState2&);
    template const MoveLogProb MoveGenerator::nextMoveWithGeneration<WHITE>(const SearchState2&);

    template const MoveLogProb MoveGenerator::nextTacticalMoveWithGeneration<BLACK>(const SearchState2&);
    template const MoveLogProb MoveGenerator::nextTacticalMoveWithGeneration<WHITE>(const SearchState2&);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
