/* IterativeLinearSolver.cc
 */
#include "osl/stat/iterativeLinearSolver.h"
#include "osl/stat/matrixMultiplier.h"
#include "osl/stat/preconditioner.h"
#include <boost/scoped_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <valarray>
#include <iostream>
#include <cmath>

#if 1
#  define cg_assert(x)
#else
#  define __MATRIX_DEBUG
#  define cg_assert(x) assert(x)
#endif

/* ------------------------------------------------------------------------- */

typedef std::valarray<double> valarray_t;

namespace osl
{
  namespace stat
  {

    template < class Matrix, class Vector, class Preconditioner, class Real >
    int 
    CG(const Matrix &A, Vector &x, const Vector &b,
       const Preconditioner &M, int &max_iter, Real &tol, bool verbose);

    template < class Matrix, class Vector, class Preconditioner, class Real >
    int 
    BiCGSTAB(const Matrix &A, Vector &x, const Vector &b,
	     const Preconditioner &M, int &max_iter, Real &tol, bool verbose);

    struct MatrixAdoptor
    {
      const MatrixMultiplier *m_mat; // acquaintance
      mutable valarray_t m_prod;
      size_t m_dim;
      MatrixAdoptor(size_t n, const MatrixMultiplier& m) : m_mat(&m), m_prod(n), m_dim(n)
      {
      }
      const valarray_t& prod(const valarray_t& x) const
      {
	cg_assert(m_dim == x.size());
	m_mat->prod(const_cast<double*>(&const_cast<valarray_t&>(x)[0]),
		    &m_prod[0]);
	return m_prod;
      }
    };
    struct PreconditionAdoptor
    {
      const Preconditioner *m_M; // acquaintance
      mutable valarray_t m_result;
      PreconditionAdoptor(size_t n, const Preconditioner *M) : m_M(M), m_result(n)
      {
      }
      void solve(const valarray_t& x, valarray_t& out) const
      {
	cg_assert(out.size() == x.size());
	m_M->solve(const_cast<double*>(&const_cast<valarray_t&>(x)[0]),
		   &out[0]);
      }
      const valarray_t& solve(const valarray_t& x) const
      {
	solve(x, m_result);
	return m_result;
      }
    };

  } // namespace stat
} // namespace osl


// CREATORS
osl::stat::
IterativeLinearSolver::IterativeLinearSolver(const MatrixMultiplier& A, const Preconditioner *M,
					     size_t max_iteration,
					     double tolerance)
  : m_A(&A), m_M(M), m_dim(A.dim()), m_max_iteration(max_iteration),
    m_tolerance(tolerance)
{
}

osl::stat::
IterativeLinearSolver::~IterativeLinearSolver()
{
}

// MANIPULATORS
int osl::stat::
IterativeLinearSolver::solve_by_CG(const valarray_t& b, valarray_t& x,
				   int *iterated, double *tol_achieved,
				   bool verbose)
{
  if (m_dim != b.size())
  {
    std::cerr << "IterativeLinearSolver: size mismatch between A " 
	      << m_dim << " and b " << b.size() << "\n";
    return 2;
  }
  if (x.size() != m_dim)
    x.resize(m_dim);

  MatrixAdoptor A_proxy(m_dim, *m_A);
  const Preconditioner *preconditioner = m_M;
  boost::scoped_ptr<IdentityPreconditioner> identity;
  if (! preconditioner)
  {
    identity.reset(new IdentityPreconditioner(m_dim));
    preconditioner = identity.get();
  }
  PreconditionAdoptor M(m_dim, preconditioner);

  double tolerance = m_tolerance;
  int max_iteration = m_max_iteration;
  const int result = CG(A_proxy, x, b, M, max_iteration, tolerance, verbose);
  if (tol_achieved)
    *tol_achieved = tolerance;
  if (iterated)
    *iterated = max_iteration;
  return result;
}

int osl::stat::
IterativeLinearSolver::solve_by_BiCGSTAB(const valarray_t& b, valarray_t& x,
					 int *iterated, double *tol_achieved,
					 bool verbose)
{
  if (m_dim != b.size())
  {
    std::cerr << "IterativeLinearSolver: size mismatch between A and b\n";
    return 2;
  }
  x.resize(m_dim);

  MatrixAdoptor A_proxy(m_dim, *m_A);
  const Preconditioner *preconditioner = m_M;
  boost::scoped_ptr<IdentityPreconditioner> identity;
  if (! preconditioner)
  {
    identity.reset(new IdentityPreconditioner(m_dim));
    preconditioner = identity.get();
  }
  PreconditionAdoptor M(m_dim, preconditioner);

  double tolerance = m_tolerance;
  int max_iteration = m_max_iteration;
  const int result
    = BiCGSTAB(A_proxy, x, b, M, max_iteration, tolerance, verbose);
  if (tol_achieved)
    *tol_achieved = tolerance;
  if (iterated)
    *iterated = max_iteration;
  return result;
}

// ACCESSORS

namespace
{
  double dot(const valarray_t& l, const valarray_t& r)
  {
    cg_assert(l.size() == r.size());
    // return std::inner_product(&l[0], &l[l.size()], &r[0], 0.0);
    double result = 0.0;
    for (size_t i=0; i<l.size(); ++i)
      result += l[i]*r[i];
    return result;
  }
  double norm(const valarray_t& l)
  {
    return sqrt(dot(l,l));
  }

} // anonymous namespace

/* ------------------------------------------------------------------------- */
// IML++ (http://math.nist.gov/iml++/)
/* ------------------------------------------------------------------------- */

//*****************************************************************
// Iterative template routine -- CG
//
// CG solves the symmetric positive definite linear
// system Ax=b using the Conjugate Gradient method.
//
// CG follows the algorithm described on p. 15 in the 
// SIAM Templates book.
//
// The return value indicates convergence within max_iter (input)
// iterations (0), or no convergence within max_iter iterations (1).
//
// Upon successful return, output arguments have the following values:
//  
//        x  --  approximate solution to Ax = b
// max_iter  --  the number of iterations performed before the
//               tolerance was reached
//      tol  --  the residual after the final iteration
//  
//*****************************************************************

template < class Matrix, class Vector, class Preconditioner, class Real >
int osl::stat::
CG(const Matrix &A, Vector &x, const Vector &b,
   const Preconditioner &M, int &max_iter, Real &tol, bool verbose)
{
  Real resid;
  Vector p(b.size()), z(b.size()), q(b.size());
  Real alpha, beta, rho, rho_1=0.0;

  Real normb = norm(b);
  Vector r = b - A.prod(x);

  if (normb == 0.0) 
    normb = 1;
  
  if ((resid = norm(r) / normb) <= tol) {
    tol = resid;
    max_iter = 0;
    return 0;
  }

  for (int i = 1; i <= max_iter; i++) {
    if (verbose)
      std::cerr << "(" << i << ") " << resid << "\n";
    if (std::isnan(resid))
      return 2;
    M.solve(r, z);
    rho = dot(r, z);
    
    if (i == 1)
      p = z;
    else {
      beta = rho / rho_1;
      p = z + beta * p;
    }
    
    q = A.prod(p);
    alpha = rho / dot(p, q);
    
    x += alpha * p;
    r -= alpha * q;
    
    if ((resid = norm(r) / normb) <= tol) {
      tol = resid;
      max_iter = i;
      return 0;     
    }

    rho_1 = rho;
  }
  
  tol = resid;
  return 1;
}

//*****************************************************************
// Iterative template routine -- BiCGSTAB
//
// BiCGSTAB solves the unsymmetric linear system Ax = b 
// using the Preconditioned BiConjugate Gradient Stabilized method
//
// BiCGSTAB follows the algorithm described on p. 27 of the 
// SIAM Templates book.
//
// The return value indicates convergence within max_iter (input)
// iterations (0), or no convergence within max_iter iterations (1).
//
// Upon successful return, output arguments have the following values:
//  
//        x  --  approximate solution to Ax = b
// max_iter  --  the number of iterations performed before the
//               tolerance was reached
//      tol  --  the residual after the final iteration
//  
//*****************************************************************


template < class Matrix, class Vector, class Preconditioner, class Real >
int osl::stat::
BiCGSTAB(const Matrix &A, Vector &x, const Vector &b,
         const Preconditioner &M, int &max_iter, Real &tol, bool verbose)
{
  Real resid;
  double rho_1, rho_2 = 0.0, alpha = 0, beta, omega;
  Vector p(b.size()), phat(b.size()), s(b.size()), shat(b.size()), t(b.size()), v(b.size());

  Real normb = norm(b);
  Vector r = b - A.prod(x);
  Vector rtilde = r;

  if (normb == 0.0)
    normb = 1;
  
  if ((resid = norm(r) / normb) <= tol) {
    tol = resid;
    max_iter = 0;
    return 0;
  }

  for (int i = 1; i <= max_iter; i++) {
    if (verbose)
      std::cerr << "(" << i << ") " << resid << "\n";
    rho_1 = dot(rtilde, r);
    if (rho_1 == 0) {
      tol = norm(r) / normb;
      return 2;
    }
    if (i == 1)
      p = r;
    else {
      beta = (rho_1/rho_2) * (alpha/omega);
      p = r + beta * (p - omega * v);
    }
    phat = M.solve(p);
    v = A.prod(phat);
    alpha = rho_1 / dot(rtilde, v);
    s = r - alpha * v;
    if ((resid = norm(s)/normb) < tol) {
      x += alpha * phat;
      tol = resid;
      max_iter = i;
      return 0;
    }
    shat = M.solve(s);
    t = A.prod(shat);
    omega = dot(t,s) / dot(t,t);
    x += alpha * phat + omega * shat;
    r = s - omega * t;

    rho_2 = rho_1;
    if ((resid = norm(r) / normb) < tol) {
      tol = resid;
      max_iter = i;
      return 0;
    }
    if (omega == 0) {
      tol = norm(r) / normb;
      return 3;
    }
  }

  tol = resid;
  return 1;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
